//-----------------------------------------------------------------------------
// from File: Billboard.cpp
//
// Desc: Example code showing how to do billboarding. The sample uses
//       billboarding to draw some trees.
//
//       Note: This implementation is for billboards that are fixed to rotate
//       about the Y-axis, which is good for things like trees. For
//       unconstrained billboards, like explosions in a flight sim, the
//       technique is the same, but the the billboards are positioned slightly
//       differently. Try using the inverse of the view matrix, TL-vertices, or
//       some other technique.
//
// Copyright (c) 1995-2000 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
{*********************************************************************
* translated to Delphi: J. Rathlev, Nov. 2001
*           rathlev@physik.uni-kiel.de, J.Rathlev@t-online.de
* required units: D3DApp, D3DFile, D3DFont, D3DUtil, DXTimer from DX8_STUFF
*                 Direct3D8, D3DX8 from Jedi-DirectX8
*   D3App, D3DFile, D3DUtil, D3DX8 and Direct3D8 are revised versions
* required DLL:  D3DX8ab.dll
* required media files:
*   skybox2.x, skybox_top.bmp, skybox_right.bmp, skybox_left.bmp,
*   skybox_front.bmp, skybox_back.bmp, skybox_bottom.bmp
*   seafloor.x, seafloor.bmp, Tree02S.tga, Tree35S.tga, Tree01S.tga
*   from Win-SDk samples
*   set "DefMediaPath" to your local path with these files
*
*********************************************************************}

unit BbMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Direct3D8, d3dx8, D3DApp, D3DFile, D3DFont, D3DUtil;

const
  DefClientWidth = 640;
  DefClientHeight = 480;
  DefMediaPath = 'media\';
  NumTrees = 500;
  NumTreeTextures = 3;
  D3DFVF_TREEVERTEX = D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1;

  g_strTreeTextures : array [0..NumTreeTextures-1] of string =
    ('Tree02S.tga','Tree35S.tga','Tree01S.tga');

  Radius = 30;

type
  TVertex = packed record
    x,y,z,tu,tv : single;
    end;

  MVertex = array of TVertex;
  PVertex = ^MVertex;

  // Custom vertex type for the trees
  TTREEVERTEX = packed record
    p     : TD3DXVECTOR3;      // Vertex position
    color : dword;             // Vertex color
    tu,tv : single;            // Vertex texture coordinates
    end;

  MTreeVertex = array of TTreeVertex;

  //-----------------------------------------------------------------------------
  // Name: Tree
  // Desc: Simple structure to hold data for rendering a tree
  //-----------------------------------------------------------------------------
  TTree = record
    v     : array [0..3] of TTREEVERTEX;  // Four corners of billboard quad
    vPos  : TD3DXVECTOR3;                 // Origin of tree
    dwTreeTexture,             // Which texture map to use
    dwOffset  : integer;       // Offset into vertex buffer of tree's vertices
    end;

  TTreeArray = array [0..NumTrees-1] of TTree;
      
  TBillBoardApp = class(TD3DApplication)
  private
    TreeVB : IDirect3DVertexBuffer8;  // Vertex buffer for rendering a tree
    TreeTextures : array of IDirect3DTexture8; // Tree images
    matBillboardMatrix : TD3DXMATRIX; // Used for billboard orientation
    Trees  : TTreeArray;           // Array of tree info
    Font : TD3DFont;              // Font for drawing text
    Terrain : TD3DMesh;           // Terrain object
    SkyBox  : TD3DMesh;           // Skybox background object
    MediaPath : string;

    constructor Create (AOwner : TComponent; AMediapath : string);
    function TreeSortCB(const arg1,arg2) : boolean;
//    procedure DrawBackground;
    procedure DrawTrees;
    function ConfirmDevice(pCaps : TD3DCAPS8;
           dwBehavior : LongWord ; Format : TD3DFORMAT) : HResult; override;
    function OneTimeSceneInit : HRESULT; override;
    function InitDeviceObjects : HRESULT; override;
    function RestoreDeviceObjects : HRESULT; override;
    function FrameMove : HRESULT; override;
    function Render : HRESULT; override;
    function InvalidateDeviceObjects : HRESULT; override;
    function DeleteDeviceObjects : HRESULT; override;
    end;

  TForm1 = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private-Deklarationen }
    BillBoardApp : TBillBoardApp;
    MediaPath : string;
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;
  gvdir : TD3DXVECTOR3;

implementation

{$R *.DFM}

function HeightField(x,y : single) : single;
begin
  Result:=9*(cos(x/20+0.2)*cos(y/15-0.2)+1.0);
  end;

constructor TBillBoardApp.Create (AOwner : TComponent; AMediapath : string);
var
  i : integer;
begin
  inherited Create (AOwner);
  Mediapath:=AMediapath;
  strWindowTitle:='Billboard: D3D Billboarding Example';
  bUseDepthBuffer:= TRUE;
  TreeVB:=nil;
  Font:=TD3DFont.Create('Arial',12,D3DFONT_BOLD);
  SkyBox:=TD3DMesh.D3DMesh('SkyBox',AMediapath);
  Terrain:=TD3DMesh.D3DMesh('SeaFloor',AMediapath);
  // Creates new mesh and texture objects based on the total
  // number of materials for the mesh.
  try
    SetLength(TreeTextures,NumTreeTextures);
  except
    exit;
    end;
  for i:=0 to NumTreeTextures-1 do TreeTextures[i]:=nil;
  end;

//-----------------------------------------------------------------------------
// Name: TreeSortCB()
// Desc: Callback function for sorting trees in back-to-front order
//-----------------------------------------------------------------------------
function TBillBoardApp.TreeSortCB (const arg1,arg2) : boolean;
var
  d1,d2 : single;
begin
  d1 := TTree(arg1).vPos.x * gvDir.x + TTree(arg1).vPos.z * gvDir.z;
  d2 := TTree(arg2).vPos.x * gvDir.x + TTree(arg2).vPos.z * gvDir.z;
  result:=(d1>d2)
  end;

//-----------------------------------------------------------------------------
// Name: DrawTrees()
// Desc:
//-----------------------------------------------------------------------------
procedure TBillBoardApp.DrawTrees;
var
  i : integer;
  matWorld : TD3DXMATRIX;
begin
  // Set diffuse blending for alpha set in vertices.
  with d3dDevice do begin
    SetRenderState( D3DRS_ALPHABLENDENABLE,   1);
    SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA );
    SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
    end;
  // Enable alpha testing (skips pixels with less than a certain alpha.)
  if(d3dCaps.AlphaCmpCaps and D3DPCMPCAPS_GREATEREQUAL <>0) then with d3dDevice do begin
    SetRenderState( D3DRS_ALPHATESTENABLE, 1);
    SetRenderState( D3DRS_ALPHAREF,        8 );
    SetRenderState( D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL );
    end;

  // Loop through and render all trees
  with d3dDevice do begin
    SetStreamSource( 0, TreeVB, sizeof(TTREEVERTEX) );
    SetVertexShader( D3DFVF_TREEVERTEX );
    for i:=0 to NumTrees-1 do with Trees[i] do begin
      // Set the tree texture
      SetTexture( 0, TreeTextures[dwTreeTexture] );

      // Translate the billboard into place
      matBillboardMatrix._41:= vPos.x;
      matBillboardMatrix._42:= vPos.y;
      matBillboardMatrix._43:= vPos.z;
      SetTransform( D3DTS_WORLD,matBillboardMatrix );

      // Render the billboard
      DrawPrimitive( D3DPT_TRIANGLESTRIP,dwOffset, 2 );
      end;

    // Restore state
    D3DXMatrixIdentity(matWorld );
    SetTransform( D3DTS_WORLD, matWorld );
    SetRenderState( D3DRS_ALPHATESTENABLE,0);
    SetRenderState( D3DRS_ALPHABLENDENABLE,0);
    end;
  end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function TBillBoardApp.ConfirmDevice(pCaps : TD3DCAPS8;
                                     dwBehavior : longword;
                                     Format : TD3DFORMAT) : HResult;
begin
  Result:=S_OK;
  // GetTransform doesn't work on PUREDEVICE
  if(dwBehavior and D3DCREATE_PUREDEVICE <>0) then result:=E_FAIL
  else begin
    // This sample uses alpha textures and/or straight alpha. Make sure the
    // device supports them
    if( pCaps.TextureCaps and D3DPTEXTURECAPS_ALPHAPALETTE <>0) then exit;
    if( pCaps.TextureCaps and D3DPTEXTURECAPS_ALPHA <>0) then exit;
    result:=E_FAIL;
    end;
  end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function TBillBoardApp.OneTimeSceneInit : HRESULT;
var
  dwColor,
  r,g,b,i        : integer;
  fWidth,fHeight,
  fTheta,fRadius : single;
begin
  // Initialize the tree data
  for I:=0 to NumTrees-1 do with Trees[i] do begin
   // Position the trees randomly
    fTheta:=2.0*D3DX_PI*random;
    fRadius:=25.0+55.0*random;
    vPos.x := fRadius * sin(fTheta);
    vPos.z := fRadius * cos(fTheta);
    vPos.y := HeightField( vPos.x, vPos.z );

    // Size the trees randomly
    fWidth:= 1.0 + 0.2 * (random-random);
    fHeight:= 1.4 + 0.4 * (random-random);

    // Each tree is a random color between red and green
    r:= (255-190) + random(190);
    g:= (255-190) + random(190);
    b:= 0;
    dwColor:= $ff000000 + (r shl 16) + (g shl 8) + b;

    v[0].p    := D3DXVECTOR3(-fWidth, 0*fHeight, 0.0);
    v[0].color:= dwColor;
    v[0].tu   := 0.0;   v[0].tv:= 1.0;
    v[1].p    := D3DXVECTOR3(-fWidth, 2*fHeight, 0.0);
    v[1].color:= dwColor;
    v[1].tu   := 0.0;   v[1].tv:= 0.0;
    v[2].p    := D3DXVECTOR3( fWidth, 0*fHeight, 0.0);
    v[2].color:= dwColor;
    v[2].tu   := 1.0;   v[2].tv:= 1.0;
    v[3].p    := D3DXVECTOR3( fWidth, 2*fHeight, 0.0);
    v[3].color:= dwColor;
    v[3].tu   := 1.0;   v[3].tv:= 0.0;

    // Pick a random texture for the tree
    dwTreeTexture:=random(NumTreeTextures);
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function TBillBoardApp.FrameMove : HRESULT;
var
  vUpVec,vEyePt,vLookatPt : TD3DXVECTOR3;
  vDir : TD3DXVECTOR3;
  matView : TD3DXMatrix;
begin
  // Get the eye and lookat points from the camera's path
  with vUpVec do begin
    x:=0.0; y:=1.0; z:=0.0;
    end;
  with vEyePt do begin
    x:= Radius*cos( 0.8 * ( dwTime/1000 ) );
    z:= Radius*sin( 0.8 * ( dwTime/1000 ) );
    y:= 4 + HeightField( x, z );
    end;
  with vLookatPt do begin
    x:= Radius*cos( 0.8 * ( dwTime/1000 + 0.5 ) );
    z:= Radius*sin( 0.8 * ( dwTime/1000 + 0.5 ) );
    y:= vEyePt.y - 1.0;
    end;

  // Set the app view matrix for normal viewing and save for render (see below)
  D3DXMatrixLookAtLH(matView,vEyePt,vLookatPt,vUpVec );
  d3dDevice.SetTransform( D3DTS_VIEW,matView );

  // Set up a rotation matrix to orient the billboard towards the camera.
  vDir:= VectorSub(vLookatPt,vEyePt);
  gvdir:=vdir;
  if( vDir.x > 0.0) then
      D3DXMatrixRotationY(matBillboardMatrix, -arctan(vDir.z/vDir.x)+D3DX_PI/2 )
  else
      D3DXMatrixRotationY(matBillboardMatrix, -arctan(vDir.z/vDir.x)-D3DX_PI/2 );

  // Sort trees in back-to-front order
  QuickSort(Trees,NumTrees,sizeof(TTree),TreeSortCB);
  Result := S_OK;
  end;

//------------------------------------------7-----------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function TBillBoardApp.Render : HRESULT;
var
  matViewSave,matViewSky : TD3DXMATRIX;
begin
  // Clear the viewport
  d3dDevice.Clear( 0, nil ,D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0, 0);
  // Begin the scene
  with d3dDevice do if SUCCEEDED(BeginScene) then begin
    // Render the Skybox
    GetTransform(D3DTS_VIEW,matViewSave);
    matViewSky:= matViewSave;
    // Center view matrix for skybox and disable zbuffer
    with matViewSky do begin
      _41:= 0.0; _42:= -0.3; _43:= 0.0;
      end;
    SetTransform( D3DTS_VIEW,matViewSky);
    SetRenderState( D3DRS_ZENABLE,D3DZB_FALSE);
    // Some cards do not disable writing to Z when
    // D3DRS_ZENABLE is FALSE. So do it explicitly
    SetRenderState( D3DRS_ZWRITEENABLE, 0);

    // Render the skybox
    SkyBox.Render(d3dDevice,true,false);

    // Restore the render states
    SetTransform( D3DTS_VIEW,matViewSave);
    SetRenderState( D3DRS_ZENABLE,D3DZB_TRUE);
    SetRenderState( D3DRS_ZWRITEENABLE, 1);

    // Draw the terrain
    Terrain.Render(d3dDevice,true,true);

    // Draw the trees
    DrawTrees;

    // Output statistics
    with Font do begin
      DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), pchar(strFrameStats),0);
      DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), pchar(strDeviceStats),0);
      end;
    // End the scene.
    EndScene;
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function TBillBoardApp.InitDeviceObjects : HRESULT;
var
  dwOffset,
  dwNumVertices,
  i,j : integer;
  vt  : MTreeVertex;
  pVB : IDIRECT3DVERTEXBUFFER8;
  pVertices : MVERTEX;
begin
  // Initialize the font's internal textures
  Result:=Font.InitDeviceObjects( d3dDevice );
  if Failed(Result) then Exit;

  // Create the tree textures
  for i:=0 to NumTreeTextures-1 do begin
    if( FAILED( D3DUtil_CreateTexture( d3dDevice, pchar(MediaPath+g_strTreeTextures[i]),
                                         TreeTextures[i],D3DFMT_UNKNOWN))) then begin
      result:=D3DAPPERR_MEDIANOTFOUND; exit;
      end;
    end;
  // Create a quad for rendering each tree
  if( FAILED(d3dDevice.CreateVertexBuffer( NumTrees*4*sizeof(TTREEVERTEX),
                                             D3DUSAGE_WRITEONLY, D3DFVF_TREEVERTEX,
                                             D3DPOOL_MANAGED,TreeVB))) then begin
      result:=E_FAIL; exit;
      end;

  // Copy tree mesh data into vertexbuffer
  SetLength(vt,NumTrees*4);
  TreeVB.Lock( 0, 0, PByte(vt), 0 );
  dwOffset:=0;
  for i:= 0 to NumTrees-1 do begin
    Move(Trees[i].v, vt[dwOffset], 4*sizeof(TTREEVERTEX));
//      for j:=0 to 3 do vt^[dwOffset+j]:=Trees[i].v[j];
    Trees[i].dwOffset:= dwOffset;
    inc(dwOffset,4);
    end;
  TreeVB.Unlock;

  // Load the skybox
  if( FAILED( SkyBox.Create1(d3dDevice,'SkyBox2.x'))) then begin
    result:=D3DAPPERR_MEDIANOTFOUND; exit;
    end;

  // Load the terrain
  if( FAILED( Terrain.Create1(d3dDevice,'SeaFloor.x'))) then begin
    RESULT:=D3DAPPERR_MEDIANOTFOUND; EXIT
    end;

  // Add some "hilliness" to the terrain
  if( SUCCEEDED( Terrain.GetSysMemMesh.GetVertexBuffer(pVB))) then begin
    dwNumVertices:= Terrain.GetSysMemMesh.GetNumVertices;
    SetLength(pVertices,dwNumVertices);
    pVB.Lock( 0, 0, PByte(pVertices), 0 );
    for i:=0 to dwNumVertices-1 do
      pVertices[i].y:= HeightField(pVertices[i].x, pVertices[i].z );
    pVB.Unlock;
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function TBillBoardApp.RestoreDeviceObjects : HRESULT;
var
  matProj : TD3DXMATRIX;
  fAspect : single;
begin
  // Restore the device objects for the meshes and fonts
  Terrain.RestoreDeviceObjects( d3dDevice );
  SkyBox.RestoreDeviceObjects( d3dDevice );
  Font.RestoreDeviceObjects;

  with d3dDevice do begin
    // Set the transform matrices (view and world are updated per frame)
    fAspect:= d3dsdBackBuffer.Width/d3dsdBackBuffer.Height;
    D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0 );
    SetTransform( D3DTS_PROJECTION, matProj );

    // Set up the default texture states
    SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
    SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
    SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );
    SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
    SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
    SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
    SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_CLAMP );
    SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_CLAMP );

    SetRenderState( D3DRS_DITHERENABLE, ord(TRUE) );
    SetRenderState( D3DRS_ZENABLE,      ord(TRUE) );
    SetRenderState( D3DRS_LIGHTING,     ord(FALSE) );
    end;
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function TBillBoardApp.InvalidateDeviceObjects : HRESULT;
begin
  Terrain.InvalidateDeviceObjects();
  SkyBox.InvalidateDeviceObjects();
  Font.InvalidateDeviceObjects();
  Result := S_OK;
  end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function TBillBoardApp.DeleteDeviceObjects : HRESULT;
var
  i : integer;
begin
  Font.DeleteDeviceObjects;
  Font:=nil;
  Terrain.Destroy;
  SkyBox.Destroy;
  TreeVB:=nil;
  Result := S_OK;
  end;

{ ------------------------------------------------------------------- }
procedure TForm1.FormCreate(Sender: TObject);
begin
  ClientWidth:=DefClientWidth; ClientHeight:=DefClientHeight;
//  MediaPath:='f:\win-sdk\samples\multimedia\media\';
  MediaPath:=DefMediaPath;
  BillBoardApp:=TBillBoardApp.Create(self,MediaPath);
  Caption:=BillBoardApp.strWindowTitle;
  end;

procedure TForm1.FormShow(Sender: TObject);
begin
  // 3D Initialization
  if Failed(BillBoardApp.Init) then Close;
  end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if assigned(BillBoardApp) then BillBoardApp.Cleanup3DEnvironment;
  end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=VK_ESCAPE then close;
  end;

end.
